@@ -1,11 +1,12 @@
 """Provides Symbolic and Reference classes"""
 from abc import ABC, abstractmethod
-from typing import Optional, Any
+from typing import ClassVar, Optional, Any
 
 import varname.helpers
 
-from .utils import Expression, evaluate_expr, logger
+from .utils import Expression, evaluate_expr
 from .context import ContextBase, ContextError
+from .function import Function
 
 class Reference(Expression, ABC):
     """The Reference class, used to define how it should be evaluated
@@ -32,59 +33,65 @@ class Reference(Expression, ABC):
             f'parent={self.parent!r}, ref={self.ref!r})'
         )
 
+    def __call__(self, *args, **kwargs):
+        """Allow `f.a.b` to be a function, so that one can do `f.a.b(1)`"""
+        return Function(self, args, kwargs, False)
+
     @abstractmethod
-    def __call__(
+    def _pipda_eval(
             self,
             data: Any,
-            context: Optional[ContextBase] = None,
-            level: int = 0
+            context: Optional[ContextBase] = None
     ) -> Any:
         """Evaluate the reference according to the context"""
-        prefix = '- ' if level == 0 else '  ' * level
-        logger.debug('%sEvaluating %r with context %r.', prefix, self, context)
-
         if context is None:
+            # needs context to be evaluated
             raise ContextError(
                 f"Cannot evaluate {repr(self)} "
                 "object without a context."
             )
 
+
 class ReferenceAttr(Reference):
     """Attribute references, for example: `f.A`, `f.A.B` etc."""
+    # Whether it's a direct reference
+    direct: ClassVar[bool] = False
 
-    def __call__(
+    def _pipda_eval(
             self,
             data: Any,
-            context: Optional[ContextBase] = None,
-            level: int = 0
+            context: Optional[ContextBase] = None
     ) -> Any:
         """Evaluate the attribute references"""
-        super().__call__(data, context, level)
-        parent = evaluate_expr(self.parent, data, context, level)
+        # if we don't have a context here, assuming that
+        # we are calling `f.a.b(1)`, instead of evaluation
+        super()._pipda_eval(data, context)
+        parent = evaluate_expr(self.parent, data, context)
 
-        return context.getattr(parent, self.ref)
+        return context.getattr(parent, self.ref, self.__class__.direct)
 
 class ReferenceItem(Reference):
     """Subscript references, for example: `f['A']`, `f.A['B']` etc"""
+    direct: ClassVar[bool] = False
 
-    def __call__(
+    def _pipda_eval(
             self,
             data: Any,
-            context: Optional[ContextBase] = None,
-            level: int = 0
+            context: Optional[ContextBase] = None
     ) -> Any:
         """Evaluate the subscript references"""
-        super().__call__(data, context, level)
-        parent = evaluate_expr(self.parent, data, context, level)
-        ref = evaluate_expr(self.ref, data, context.ref, level)
-        return context.getitem(parent, ref)
+        super()._pipda_eval(data, context)
+        parent = evaluate_expr(self.parent, data, context)
+        ref = evaluate_expr(self.ref, data, context.ref)
+        return context.getitem(parent, ref, self.__class__.direct)
 
 class DirectRefAttr(ReferenceAttr):
     """The direct attribute reference, such as `f.A`"""
-
+    direct: ClassVar[bool] = True
 
 class DirectRefItem(ReferenceItem):
     """The direct attribute reference, such as `f['A']`"""
+    direct: ClassVar[bool] = True
 
 @varname.helpers.register
 class Symbolic(Expression):
@@ -98,16 +105,15 @@ class Symbolic(Expression):
 
     def __getitem__(self, item: Any) -> Any:
         """Create a DirectRefItem object"""
-        return ReferenceItem(self, item)
+        return DirectRefItem(self, item)
 
     def __repr__(self) -> str:
         return f"<Symbolic:{self.__varname__}>"
 
-    def __call__(
+    def _pipda_eval(
             self,
             data: Any,
-            context: Optional[ContextBase] = None,
-            level: int = 0
+            context: Optional[ContextBase] = None
     ) -> Any:
         """When evaluated, this should just return the data directly"""
         return data
